from .adis_block import AdisBlock
from .adis_lines import DefinitionLine

"""
An AdisFile contains multiple AdisBlocks.
"""

class AdisFile:
    def __init__(self, blocks):
        """Creates a new AdisFile

        Args:
            blocks (list[AdisBlock]): blocks that are located in this file
        """
        self.blocks = blocks

    def get_blocks(self):
        """Returns a list of AdisBlocks that are in this AdisFile

        Returns:
            list[AdisBlock]: AdisBlocks in this AdisFile
        """
        return self.blocks

    @staticmethod
    def from_lines(lines):
        """Creates a new AdisFile from the provided lines.

        Args:
            lines (list[AdisLine]): the AdisFile will be created on base of these lines

        Returns:
            AdisFile: the new AdisFile
        """
        blocks_of_lines = AdisFile.split_lines_into_blocks(lines)
        blocks = []

        for block_of_lines in blocks_of_lines:
            blocks.append(AdisBlock.from_lines(block_of_lines))

        return AdisFile(blocks)

    @staticmethod
    def from_dict(file_dict):
        """Creates a new AdisFile from the provided dict.

        Args:
            file_dict (dict): contains information about the contents of this file

        Returns:
            AdisFile: the new AdisFile
        """
        blocks = []
        for entity_number in file_dict:
            block_dict = file_dict[entity_number]
            adis_block = AdisBlock.from_dict(entity_number, block_dict)
            blocks.append(adis_block)
        return AdisFile(blocks)

    # Each block starts with a definition
    @staticmethod
    def split_lines_into_blocks(lines):
        """Splits the lines into blocks of lines.

        Args:
            lines (list[AdisLine]): list of AdisLines of this file

        Returns:
            list[list[AdisLine]]: list that contains a list of lines for each block in this \
                AdisFile
        """
        blocks = []
        current_block = []

        for line in lines:
            if type(line) == DefinitionLine:
                # start of a new block
                if len(current_block) != 0:
                    blocks.append(current_block)
                current_block = [line]
            else:
                current_block.append(line)
        
        if len(current_block) != 0:
            blocks.append(current_block)

        return blocks

    def to_dict(self):
        """Creates a dict from the AdisFile.

        Returns:
            dict: contains all data of the blocks in this file
        """
        data = {}
        for block in self.blocks:
            data[block.get_entity_number()] = block.to_dict()
        return data

    def dumps(self):
        """Creates an ADIS text from this AdisFile.

        Returns:
            string: ADIS text of this AdisFile
        """
        text = ""
        for block in self.blocks:
            text += block.dumps()

        return text

    def __repr__(self):
        return "AdisFile contains %d blocks" % len(self.blocks)



from .adis_field_definition import AdisFieldDefinition
from .adis_lines import (
    AdisLine,
    DefinitionLine,
    ValueLine
)
from .adis_value import AdisValue

"""
An AdisBlock consists of one definition and (one or) multiple data rows.
Each data row has multiple fields.
"""

class AdisBlock:
    def __init__(self, entity_number, status, field_definitions, data_rows):
        """Creates an AdisBlock object.

        Args:
            entity_number (string): Entity number of this Block (has to be a string with 6 chars)
            status (string): Status char of this block, can be H, N, S, F or D
            field_definintions (list[AdisFieldDefinition]): Field definitions
            data_rows (list[list[AdisValue]]): list of data rows, each data row is a list \
                containing AdisValue
        """
        if len(status) != 1:
            raise Exception("Status may only be one char.")
        if status not in AdisLine.status_chars:
            raise Exception("Invalid status char. Has to be one of %s."
                % AdisLine.status_chars)
        if type(entity_number) is not str or len(entity_number) != 6:
            raise Exception("""The entity number has to be a string consisting of 6 chars.
                Got \"%s\".""" % entity_number)
        self.status = status
        self.entity_number = entity_number
        self.field_definitions = field_definitions
        self.data_rows = data_rows

    def get_entity_number(self):
        """Returns the entity number of this block.

        Returns:
            string: Entity number of this block (is a string of 6 chars)
        """
        return self.entity_number

    def get_field_definitions(self):
        """Returns the field definitions of this block.

        Returns:
            list[AdisFieldDefinition]: list containing field definitions
        """
        return self.field_definitions
    
    def get_data_rows(self):
        """Returns the data rows of this block.

        Returns:
            list[list[AdisValue]]: list of data rows. Each row is a list of AdisValues
        """
        return self.data_rows

    @staticmethod
    def from_lines(lines):
        """Creates an AdisBlock from a list of AdisLines

        Args:
            lines (list[AdisLine]): list of AdisLines
        
        Returns:
            AdisBlock: new AdisBlock
        """
        status = None
        entity_number = None
        field_definitions = None
        data_rows = []
        for line in lines:
            if type(line) == DefinitionLine:
                status = line.get_status_char()
                entity_number = line.get_entity_number()
                field_definitions = line.get_field_definitions()
            if type(line) == ValueLine:
                if field_definitions is None:
                    raise Exception("Definition line is missing before value line")
                data_rows.append(line.parse(field_definitions))

        if entity_number is None or field_definitions is None:
            raise Exception("Definition is missing.")

        return AdisBlock(entity_number, status, field_definitions, data_rows)

    @staticmethod
    def from_dict(entity_number, block_dict):
        """Creates an AdisBlock form a dict.

        Args:
            entity_number (string): Entity number of the block
            block_dict (dict): containing the data rows, the field definitions and the status

        Returns:
            AdisBlock: new AdisBlock
        """
        if "status" not in block_dict:
            raise Exception("\"status\" field is missing in block dict.")
        if "definitions" not in block_dict:
            raise Exception("\"definitions\" field is missing in block dict.")
        if "data" not in block_dict:
            raise Exception("\"data\" field is missing in block dict.")

        status = block_dict["status"]
        field_definitions = []
        for definition_dict in block_dict["definitions"]:
            if type(definition_dict) is not dict:
                raise Exception("A field definition has to be a dict but got %s."
                                % type(definition_dict))
            definition = AdisFieldDefinition.from_dict(definition_dict)
            field_definitions.append(definition)

        if type(block_dict["data"]) is not list:
            raise Exception("The data of each block has to be a list. Got %s."
                            % type(block_dict["data"]))
        
        data_rows = []
        for data_row_dict in block_dict["data"]:
            if type(data_row_dict) is not dict:
                raise Exception("Each data row has to be a dict. Got %s."
                                % type(data_row_dict))
            data_row = []
            for definition in field_definitions:
                item_number = definition.get_item_number()
                if item_number in data_row_dict:
                    value = data_row_dict[item_number]
                    adis_value = AdisValue(item_number, value)
                    data_row.append(adis_value)

            data_rows.append(data_row)

        return AdisBlock(entity_number, status, field_definitions, data_rows)

    def to_dict(self):
        """Creates a dict cointaining all data of this block

        Returns:
            dict: contains the field definitions, the data rows and the status of the block
        """
        result_dict = {
            "definitions": [],
            "data": [],
            "status": self.status
        }
        for definition in self.field_definitions:
            result_dict["definitions"].append(definition.to_dict())

        for data_row in self.data_rows:
            result_dict["data"].append(self.data_row_to_dict(data_row))

        return result_dict

    def data_row_to_dict(self, data_row):
        """Turns a data row to a dict.

        Args:
            data_row (list[AdisValue]): list containing the values of the data row

        Returns:
            dict: dict where the key is the item number and the value is the actual value of the \
                field
        """
        data_row_dict = {}
        for value in data_row:
            value_as_dict = value.to_dict()
            data_row_dict[value_as_dict["item_number"]] = value_as_dict["value"]
        return data_row_dict

    def dumps_definitions(self):
        """Creats a ADIS definition line string from the field definitions.

        Returns:
            string: ADIS definition line string
        """
        text = "D" + self.status + self.entity_number
        for definition in self.field_definitions:
            text += definition.dumps()
        text += "\r\n"
        return text

    def dumps_data(self):
        """Turns the data rows into value lines.

        Returns:
            string: value lines
        """
        text = ""
        for data_row in self.data_rows:
            # create a value line
            text += "V" + self.status + self.entity_number
            data_row_dict = self.data_row_to_dict(data_row)
            for definition in self.field_definitions:
                item_number = definition.get_item_number()
                if item_number not in data_row_dict:
                    # the value of this field is undefined
                    text += definition.dumps_value(None, undefined=True)
                else:
                    value = data_row_dict[item_number]
                    text += definition.dumps_value(value)
            
            text += "\r\n"
        return text

    def dumps(self):
        """Creates an ADIS text form this block.

        Returns:
            string: ADIS text of this block
        """
        text = self.dumps_definitions()
        text += self.dumps_data()
        return text

    def __repr__(self):
        return "AdisBlock with status=%s and entity_number=%s containing %d data row(s)" \
                % (self.status, self.entity_number, len(self.data_rows))



from .adis_value import AdisValue

"""
The AdisFieldDefinition holds information about the size and the decimal places of the data fields.
"""
class AdisFieldDefinition:
    def __init__(self, item_number, field_size, decimal_digits):
        """Creates an AdisFieldDefinition.

        Args:
            item_number (string): item number of the fields
            field_size (string, int): field size in chars
            decimal_digits (string, int): number of decimal places (0 when the field does not \
                hold a text or the number has no decimal places)
        """
        if type(item_number) is not str or len(item_number) != 8:
            raise Exception("The item_number has to be a string with length = 8. Got \"%s\""
                % item_number)
        self.item_number = item_number
        
        self.field_size = int(field_size)
        if self.field_size < 1 or 99 < self.field_size:
            raise Exception("The field size has to be a number between 1 and 99. Got %d."
                % self.field_size)
        
        self.decimal_digits = int(decimal_digits)
        if self.decimal_digits < 0 or 9 < self.decimal_digits:
            raise Exception("The number of decimal digits has to be a number between 0 and 9. " \
                "Got %d." % self.decimal_digits)

    def get_item_number(self):
        """Returns the item number.

        Returns:
            string: item number
        """
        return self.item_number

    def get_field_size(self):
        """Returns the field size.

        Returns:
            int: field size
        """
        return self.field_size

    def get_decimal_digits(self):
        """Returns the number of decimal digits.

        Returns:
            int: number of decimal digits
        """
        return self.decimal_digits

    def parse_field_at_position(self, raw_text, position):
        """Parses 

        Args:
            raw_text (string): raw ADIS file line
            position (int): position where to start to parse the field from

        Returns:
            AdisValue: AdisValue of the parsed field, or None if the value of the field is \
                undefined
        """
        # return a string when decimal_digits is 0 / otherwise float
        item_number = self.item_number
        value = raw_text[position:position + self.field_size]

        value_size = len(value)
        if value_size != 0 and value_size != self.field_size:
            raise Exception("Expected field size of %d chars or an empty field, but got " \
                "field size of %d chars." % (self.field_size, value_size))
        elif value_size == 0:
            value = None

        if value is not None and self.only_contains_char(value, "?"):   # null value field
            value = None

        if value is not None and self.only_contains_char(value, "|"):   # undefined DDI number
            return None         # no value will be created for this field

        # handle case where it's a decimal number
        if value is not None and self.decimal_digits != 0:
            value = float(value)
            value /= 10**self.decimal_digits

        return AdisValue(item_number, value)

    def only_contains_char(self, text, allowed_char):
        """Checks if the provided text only contains the provided char.

        Args:
            text (string): text to check
            allowed_char (string): char that may be in the text

        Returns:
            boolean: True if the text only contains the given char, otherwise False
        """
        for char in text:
            if char != allowed_char:
                return False
        return True

    def to_dict(self):
        """Creates a dict that contains all information from this definition.

        Returns:
            dict: contains the item_number, field_size and decimal_digits
        """
        return {
            "item_number": self.item_number,
            "field_size": self.field_size,
            "decimal_digits": self.decimal_digits
        }

    def dumps(self):
        """Creates a string from this definition that can directly be a part of the defintion \
            line in an ADIS file

        Returns:
            string: definition text
        """
        text = self.item_number
        text += self.number_to_str_with_window_size(self.field_size, 2)
        text += str(self.decimal_digits)    # only 1 char window
        return text

    def dumps_value(self, value, undefined=False):
        """Dumps the provided AdisValue to a string.

        Args:
            value (AdisValue): value that should be turned to a string
            undefined (bool, optional): Whether the value is undefined or not. Defaults to False.

        Returns:
            string: string that holds the given AdisValue in the correct way
        """
        if undefined:
            return self.field_size * "|"
        if value is None:
            return self.field_size * "?"
        elif type(value) is str:
            if self.field_size < len(value):
                raise Exception("value \"%s\" is too long for this field." % value)
            
            # fill space on the right with " "'s
            missing_chars_number = self.field_size - len(value)
            return value + missing_chars_number * " "
        else:
            text = str(value)
            if self.decimal_digits != 0 or type(value) is float:
                decimal_dot_position = text.find(".") + 1
                decimal_dot_position_from_behind = len(text) - decimal_dot_position
                text = text.replace(".", "")
                
                expected_decimal_dot_position_from_behind = self.decimal_digits

                if expected_decimal_dot_position_from_behind > decimal_dot_position_from_behind:
                    decimal_places_to_add = expected_decimal_dot_position_from_behind \
                                                - decimal_dot_position_from_behind
                    text += decimal_places_to_add * "0"
                elif expected_decimal_dot_position_from_behind < decimal_dot_position_from_behind:
                    decimal_places_to_remove = decimal_dot_position_from_behind \
                                                - expected_decimal_dot_position_from_behind
                    text = text[:len(text) - decimal_places_to_remove]

            text_length = len(text)
            if self.field_size < text_length:
                raise Exception(f"Number {value} is too large for this field.")

            missing_chars_number = self.field_size - text_length
            text = " " * missing_chars_number + text
            return text

    def number_to_str_with_window_size(self, number, window_size):
        """Creates a string with a specific length from a number. All chars that are not used get \
            filled up with "0"s

        Args:
            number (int): number that should be turned to string
            window_size (int): how long the string should be

        Returns:
            string: text with specified length that contains the number
        """
        result_str = str(number)
        missing_chars_number = window_size - len(result_str)
        result_str = missing_chars_number * "0" + result_str
        return result_str

    @staticmethod
    def from_dict(definition_dict):
        """Creates an AdisFieldDefinition from a dict.

        Args:
            definition_dict (dict): dict containing the item number, the field size and the \
                decimal digits of field

        Returns:
            AdisFieldDefinition: new AdisFieldDefinition
        """
        item_number = definition_dict["item_number"]
        field_size = definition_dict["field_size"]
        decimal_digits = definition_dict["decimal_digits"]
        return AdisFieldDefinition(item_number, field_size, decimal_digits)

    def __repr__(self):
        return "AdisFieldDefinition: item_number=%s, field_size=%d, decimal_digits=%d" \
            % (self.item_number, self.field_size, self.decimal_digits)



import json
from .adis_file import AdisFile
from .adis_lines import (
    AdisLine,
    EndOfLogicalFileLine,
    PhysicalEndOfFileLine
)
from .adis_value import AdisValue

def split_lines(raw_input_text):
    """Splits the provided text into lines. Lines are splitted at "\n" and "\r" chars get \
        removed as well.

    Args:
        raw_input_text (string): ADIS text

    Returns:
        list: lines of the provided ADIS text
    """
    # lines have to end with "\r\n" but we also accept lines that only end with "\n"
    lines = []
    for raw_line in raw_input_text.split("\n"):
        lines.append(raw_line.replace("\r", ""))    # remove \r
    return lines


class Adis:
    def __init__(self, adis_files):
        """Creates an Adis object based on the logical ADIS files

        Args:
            adis_files (list[AdisFile]): List of logical ADIS files
        """
        self.files = adis_files

    def get_files(self):
        """Returns a list containing the logical AdisFiles

        Returns:
            list[AdisFile]: list containing the logical AdisFiles
        """
        return self.files

    @staticmethod
    def parse(text):
        """This method parses the provided ADIS text into an Adis object.

        Args:
            text (string): ADIS file content

        Returns:
            Adis: Adis object created from the provided ADIS text
        """
        raw_lines = split_lines(text)

        adis_files = []

        lines_for_file = []
        for raw_line in raw_lines:
            if raw_line == "":
                continue

            line = AdisLine.parse_line(raw_line)
            if type(line) == EndOfLogicalFileLine or type(line) == PhysicalEndOfFileLine:
                adis_files.append(AdisFile.from_lines(lines_for_file))
                lines_for_file = []
            else:
                lines_for_file.append(line)

        return Adis(adis_files)

    @staticmethod
    def parse_from_file(path_to_file):
        """This method parses the given ADIS file to an Adis object.

        Args:
            path_to_file (string): Path to the ADIS file

        Returns:
            Adis: Adis object created from the provided ADIS file
        """
        with open(path_to_file, "r") as input_file:
            text = input_file.read()
        return Adis.parse(text)

    def get_list(self, strip_string_values=True):
        """Returns a list containing of the logical ADIS files and their contents. \
            The returned list only contains builtin types.

        Args:
            strip_string_values (bool, optional, by default True): Whether string \
                values should be stripped or not.

        Returns:
            list: containing the logical adis files and their contents as builtin types.
        """

        AdisValue.strip_string_values = strip_string_values
        list_of_files = []
        for adis_file in self.files:
            list_of_files.append(adis_file.to_dict())
        return list_of_files

    def to_json(self, strip_string_values=True, mapping_dict: dict=None):
        """Creates a json from the Adis object.

        Args:
            strip_string_values (bool, optional, by default True): Whether string \
                values should be stripped or not.
            mapping_dict (dict): Optional dictionary of mapping values \
                for entity numbers (e.g. {"0080004": "Betriebsnummer"})

        Returns:
            string: Adis as json
        """
        if mapping_dict is None or type(mapping_dict) != dict:
            return json.dumps(self.get_list(strip_string_values))
        else:
            list_of_adis = self.get_list(strip_string_values)
            return json.dumps(self.add_string_value(list_of_adis, mapping_dict))

    @staticmethod
    def from_json(json_text):
        """Creates an Adis object based on the provided json.

        Args:
            json_text (string): json text. Note that the json has to have a specific structure. \
                Take a look at the README for more information.

        Returns:
            Adis: Adis object
        """
        data = json.loads(json_text)
        if type(data) is not list:
            raise Exception("""The root element of the JSON has to be a list. Got %s."""
                            % type(data))
        adis_files = []
        for file_dict in data:
            if type(file_dict) is not dict:
                raise Exception("Expecting a dict but got %s." % type(file_dict))
            adis_files.append(AdisFile.from_dict(file_dict))

        return Adis(adis_files)

    @staticmethod
    def from_json_file(path_to_json_file):
        """Creates an Adis object based on the provided json file.

        Args:
            path_to_json_file (string): Path to the json file. Note that the json has to have a \
                specific structure. Take a look at the README for more information.

        Returns:
            Adis: Adis object
        """
        with open(path_to_json_file) as input_file:
            raw_content = input_file.read()
        return Adis.from_json(raw_content)

    def dumps(self):
        """Creates an ADIS text

        Returns:
            string: ADIS text
        """
        file_texts = []
        for adis_file in self.files:
            file_texts.append(adis_file.dumps())
        
        end_of_logical_file_line = "EN\r\n"
        text = end_of_logical_file_line.join(file_texts)

        text += "ZN\r\n"    # physical end of file
        return text

    def __repr__(self):
        return """Adis containing %d Adis-files""" % len(self.files)

    @staticmethod
    def add_string_value(obj: list, entity_data_dict: dict):
        """ Adds string value to definition list

        Args:
            obj (list): containing the logical adis files and their contents as builtin types without string values
            entity_data_dict (dict): dictionary of mapping values for \
             item numbers (e.g. {"0080004": "Betriebsnummer"})

        Returns:
             list: containing the logical adis files and their contents as builtin types with string values

        """
        for o in obj:
            for defs in o:
                for items in obj[0].get(defs).get("definitions"):
                    x = entity_data_dict.get(items['item_number'][2:], items['item_number'][2:])
                    items['item_name'] = x
        return obj



class AdisValue:
    strip_string_values = True

    def __init__(self, item_number, value):
        """Creates a AdisValue from the item number and a value of a field.

        Args:
            item_number (string): item number of the field
            value (None, string, int, float): value of the field
        """
        self.item_number = item_number
        self.value = value

    def to_dict(self):
        """Returns the AdisValue as a dict.

        Returns:
            dict: contains item_number and value
        """

        result_dict = {
            "item_number": self.item_number
        }

        if AdisValue.strip_string_values and isinstance(self.value, str):
            result_dict["value"] = self.value.strip()
        else:
            result_dict["value"] = self.value

        return result_dict

    def __repr__(self):
        return "AdisValue: item_number=%s, value=%s" % (self.item_number, str(self.value))



from .adis_field_definition import AdisFieldDefinition

class AdisLine:
    status_chars = {
        "H": "header",
        "N": "normal",
        "S": "synchronisation",
        "F": "faulty",
        "D": "deletion"
    }

    # takes the line without the first char
    def __init__(self, line):
        """Creates an AdisLine.

        Args:
            line (string): the raw line from the ADIS file
        """
        self.line = line
        self.line_type_char = line[0]
        self.status_char = line[1]

        self.status = AdisLine.status_chars[self.status_char]

        if not self.status_allowed(self.status):
            raise Exception(
                f"Linetype {self.line_type} may not have status {self.status}")

    def status_allowed(self, status):
        """Returns whether the given status is allowed for this AdisLine or not.

        Args:
            status (string): status char

        Returns:
            boolean: True if this status is allowed for this line, otherwise False
        """
        return status in self.allowed_statuses

    def get_type_char(self):
        """Returns the type char of this AdisLine.

        Returns:
            string: type char of this AdisLine
        """
        return self.line_type_char

    def get_status_char(self):
        """Returns the status char of this AdisLine.

        Returns:
            string: status char of this AdisLine
        """
        return self.status_char

    def __repr__(self):
        return "%s status: %s, line: %s" % (self.line_type, self.status, self.line)
    
    @staticmethod
    def parse_line(line):
        """Creates an AdisLine from the given line.

        Args:
            line (string): line from the ADIS file

        Returns:
            AdisLine: new AdisLine
        """
        line_types = {
            "D": DefinitionLine,
            "V": ValueLine,
            "E": EndOfLogicalFileLine,
            "C": CommentLine,
            "Z": PhysicalEndOfFileLine,
            "T": PhysicalEndOfFileLine
        }

        line_type = line[0]
        return line_types[line_type](line)


class DefinitionLine(AdisLine):
    def __init__(self, line):
        """Creates a DefinitionLine

        Args:
            line (string): line from ADIS file
        """
        self.line_type = "Definition"
        self.allowed_statuses = [
            "header", "normal", "synchronisation", "faulty", "deletion"
        ]
        super().__init__(line)

        self.entity_number = self.line[2:8]

        field_definitions_text = self.line[8:]
        field_definition_blocks = self.split_field_definitions_text(field_definitions_text)

        self.field_definitions = []
        for field_definition_block in field_definition_blocks:
            self.field_definitions.append(self.create_field_definition(field_definition_block))

    def get_entity_number(self):
        """Returns the entity number.

        Returns:
            string: entity number if this line
        """
        return self.entity_number

    def get_field_definitions(self):
        """returns the field definitions.

        Returns:
            list[AdisFieldDefinition]: list of field definitions
        """
        return self.field_definitions

    def split_field_definitions_text(self, field_definitions_text):
        """Splits the part of an ADIS file line that contains the field definitions into parts. \
            (Each field definition is one part)

        Args:
            field_definitions_text (string): part of a raw ADIS line that holds the field \
                definitions

        Returns:
            list[string]: list that holds the different field definition strings
        """
        definition_block_length = 11    # each field definition consists of 11 chars
        if len(field_definitions_text) % definition_block_length != 0:
            raise Exception("Length of definitions text is %d but it has to be a multiple of %d"
                % (len(field_definitions_text), definition_block_length))
        definition_blocks = []
        number_of_field_definitions = int(len(field_definitions_text) / definition_block_length)
        for block_count in range(number_of_field_definitions):
            start_index = block_count * definition_block_length
            end_index = start_index + definition_block_length
            definition_blocks.append(field_definitions_text[start_index:end_index])
        return definition_blocks

    def create_field_definition(self, field_definition_block):
        """Creates an AdisFieldDefinition from the raw definition block from the ADIS line.

        Args:
            field_definition_block (string): part of the raw ADIS line that holds a field \
                definition

        Returns:
            AdisFieldDefinition: new AdisFieldDefinition
        """
        item_number = field_definition_block[0:8]
        field_size = field_definition_block[8:10]
        decimal_digits = field_definition_block[10]
        return AdisFieldDefinition(item_number, field_size, decimal_digits)


class ValueLine(AdisLine):
    def __init__(self, line):
        """Creates a ValueLine.

        Args:
            line (string): raw line from an ADIS file
        """
        self.line_type = "Value"
        self.allowed_statuses = [
            "header", "normal", "synchronisation", "faulty", "deletion"
        ]
        super().__init__(line)

        self.entity_number = self.line[2:8]
        self.raw_items = self.line[8:]

    def get_entity_number(self):
        """Returns the entity number.

        Returns:
            string: entity number if this line
        """
        return self.entity_number

    def parse(self, field_definitions):
        """Parses a list of AdisValues from a raw ADIS file value line using the provided \
            list of field definitions.

        Args:
            field_definitions (list[AdisFieldDefinition]): The field definitions for the fields \
                in this line.

        Returns:
            list[AdisValue]: AdisValues parsed from this line
        """
        expected_raw_items_length = 0
        for field_definition in field_definitions:
            expected_raw_items_length += field_definition.get_field_size()
        
        if len(self.raw_items) != expected_raw_items_length:
            expected_raw_items_length_when_last_item_empty = \
                expected_raw_items_length - field_definitions[-1].get_field_size()
            if len(self.raw_items) < expected_raw_items_length_when_last_item_empty:
                raise Exception("Expecting an item text length of %d chars or %d chars, " \
                    "but got %d chars."
                    % (expected_raw_items_length, 
                        expected_raw_items_length_when_last_item_empty,
                        len(self.raw_items)))

        values = []
        current_position = 0
        for field_definition in field_definitions:
            field_size = field_definition.get_field_size()
            value = field_definition.parse_field_at_position(self.raw_items, current_position)
            if value is not None:
                values.append(value)
            current_position += field_size

        return values


class EndOfLogicalFileLine(AdisLine):
    def __init__(self, line):
        """Creates an EndOfLogicalFileLine. Note that each physical ADIS file can contain \
            multiple logical files.

        Args:
            line (string): raw line from an ADIS file
        """
        self.line_type = "End of logical file"
        self.allowed_statuses = ["normal"]
        super().__init__(line)


class CommentLine(AdisLine):
    def __init__(self, line):
        """Creates a CommentLine

        Args:
            line (string): raw line from an ADIS file
        """
        self.line_type = "Comment"
        self.allowed_statuses = [
            "header", "normal", "synchronisation", "faulty", "deletion"
        ]
        super().__init__(line)
        self.comment = self.line[2:]
    
    def get_comment(self):
        """Returns the comment.

        Returns:
            string: comment
        """
        return self.comment


class PhysicalEndOfFileLine(AdisLine):
    def __init__(self, line):
        """Creates a PhysicalEndOfFileLine.

        Args:
            line (string): raw line from an ADIS file
        """
        self.line_type = "Physical end of file"
        self.allowed_statuses = ["normal"]
        super().__init__(line)
